import inspect

from typing import Type

from rest_framework.exceptions import APIException


class AllowClassBehaviorAsFunction:
    """
    This class should be used when the inherited class instance should be 'called' as a function.
    It's very useful when it's necessary to initiate a PermissionClass and then add it on `permission_class`.
    """

    def __call__(self):
        """
        The __call__ method enables classes to be writen where the instances behave like functions and can be called
        like a function. When the instance is called as a function; if this method is defined,
        x(arg1, arg2, ...) is a shorthand for x.__call__(arg1, arg2, ...).

        Google Recaptcha Permission Example:
        This method allows to use this Permission Class with a custom response without being forced to use
        "get_permissions".

        Without this method, is necessary to do something like this:
            def get_permissions(self) -> List:
                exc = SomeException(status.HTTP_401_UNAUTHORIZED, {"details": "Some message"})
                return [GoogleRecaptchaPermission(exc)]

        With this method, one can achieve the same result using something like this:
            permission_classes = [
                GoogleRecaptchaPermission(SomeException(status.HTTP_401_UNAUTHORIZED, {"details": "Some message"}))
            ]
        """

        return self


def is_exception_related_to_api_exception(
    exception_in_case_of_failed_verification: APIException | Type[APIException],
) -> bool:
    """
    Checks if the received exception is an instance or a subclass of APIException.
    The "isinstance" handles cases like this:
        class MyException(APIException):
            pass
        class MyViewset(viewsets.ModelViewSet):
            permission_class = [GoogleRecaptchaPermission(MyException())]
    While "issubclass" handles cases like this:
        class MyException(APIException):
            pass
        class MyViewset(viewsets.ModelViewSet):
            permission_class = [GoogleRecaptchaPermission(MyException)]
    """

    exc = exception_in_case_of_failed_verification
    if not exc:
        return False

    exception_class = exc if inspect.isclass(exc) else type(exc)
    # isinstance deals with initialized exceptions
    # issubclass deals with not initialized exceptions
    return isinstance(exc, APIException) or issubclass(exception_class, APIException)



import re

from typing import Callable
from typing import List


def validate_format_email(email: str) -> bool:
    """
    Check if the given string has a valid format e-mail
    """
    regex = r"\b^[A-Za-z0-9](([_\.\-]?[a-zA-Z0-9\_]+)*)@([A-Za-z0-9]+)(([\.\-]?[a-zA-Z0-9]+)*)\.([A-Za-z]{2,})$\b"
    return bool(re.match(regex, email))


def perform_callback_function_validators(functions_to_validate_email: List[Callable], email: str) -> bool:
    """
    Returns if the e-mail is valid by `functions_to_validate_email`
        Parameters:
            `functions_to_validate_email (List[Callable])`: a list of functions to validate the e-mail
            `email (str)`: the e-mail that will be validated
    """
    for check in functions_to_validate_email:
        checkResult = check(email)
        if not checkResult:
            return False
    return True



class MissingRequiredConfiguration(Exception):
    pass


class IncorrectTypePermissionConfiguration(Exception):
    pass


class ApplicationSettingsNotFound(Exception):
    """
    Exception that indicates the application settings was not found through known methods.
    """


class RequestIDModuleNotFound(Exception):
    """
    Exception that indicates the request id module was not found through known methods.
    """


class DecoratorWrongUsage(Exception):
    """
    Generic exception when a decorator is used wrongly.
    """


class EmailValidationError(Exception):
    """
    Exception that indicates Everest is down and the default validation methods will not be used.
    """



"""
HTTP utils module
"""

import functools
import logging

from contextlib import contextmanager
from importlib import import_module
from typing import Any
from typing import Generator
from typing import Optional

from requests import Response
from requests import Session
from requests.adapters import HTTPAdapter
from urllib3 import Retry


def _log_response(response: Response, *args: Any, **kwargs: Any) -> None:
    logging.debug("Request to %s returned status code %d", response.url, response.status_code)


def _check_for_errors(response: Response, *args: Any, **kwargs: Any) -> None:
    response.raise_for_status()


@contextmanager
def request(total: int = 3, backoff_factor: float = 0.1, **kwargs: Any) -> Generator[Session, None, None]:
    """
    Generate a requests session that allows retries and raises HTTP errors (status code >= 400).
    Uses the same arguments as the class Retry from urllib3
    """
    settings = import_module("jsm_user_services.settings")
    JSM_USER_SERVICE_HTTP_TIMEOUT = float(getattr(settings, "JSM_USER_SERVICE_HTTP_TIMEOUT", 30))

    session = Session()
    session.hooks.update(response=[_log_response, _check_for_errors])
    max_retries = Retry(total=total, backoff_factor=backoff_factor, **kwargs)
    adapter = HTTPAdapter(max_retries=max_retries)
    session.mount("https://", adapter)
    session.mount("http://", adapter)

    session.request = functools.partial(session.request, timeout=JSM_USER_SERVICE_HTTP_TIMEOUT)  # type: ignore
    try:
        yield session
    finally:
        session.close()


def get_response_body(response: Response) -> Any:
    """
    Deserialize response
    """
    try:
        response_body = response.json()
    except ValueError:
        response_body = response.text
    return response_body


def convert_header_to_meta_key(header: Optional[str]) -> Optional[str]:
    """
    Django Http Request has a dictionary called META with the request headers.
    But it replaces the header name following the rules:
      - all characters to uppercase
      - replacing any hyphens with underscores
      - adding an HTTP_ prefix to the name.
    So, for example, a header called X-Bender would be mapped to the META key HTTP_X_BENDER.
    """
    if header is None:
        return None

    return f"HTTP_{header.replace('-', '_')}".upper()



import logging

from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import TypedDict

from requests import Response

from jsm_user_services.exception import EmailValidationError
from jsm_user_services.support.email_utils import perform_callback_function_validators
from jsm_user_services.support.email_utils import validate_format_email
from jsm_user_services.support.http_utils import get_response_body
from jsm_user_services.support.http_utils import request
from jsm_user_services.support.settings_utils import get_from_settings_or_raise_missing_config

logger = logging.getLogger(__name__)


DEFAULT_VALIDATORS_FUNCTION: List[Callable] = [validate_format_email]


class ValidationResult(TypedDict):
    is_valid: bool
    external_api_validation: bool


def perform_email_validation(
    email: str,
    use_callback_function_validator: bool = True,
    functions_to_validate_email: Optional[List[Callable]] = None,
) -> ValidationResult:
    """
    Parameters:
        `email (str)`: the e-mail that will be validated
        `use_callback_function_validator (bool)`: indicates if some validators must be considered if the `status_code`
        is different from 200 (success)
        `functions_to_validate_email (List[Callable])`: list of validators that will be considered if the `status_code`
        is different from 200 (success) and the `use_callback_function_validator` is `False`

    Returns:
        A dict containing two values:
        - `is_valid (bool)`: indicates if the email string is valid
        - `external_api_validation (bool)`: indicates if the email validation was made by Everest or locally

    Response success example:
        {
            "meta": {},
            "results": {
                "category": "valid",
                "status": "valid",
                "name": "Valid",
                "definition": "The email has a valid account associated with it.",
                "reasons": [],
                "risk": "low",
                "recommendation": "send",
                "address": "ayron41@gmail.com",
                "diagnostics": {
                    "role_address": false,
                    "disposable": false,
                    "typo": false
                }
            }
        }
    """
    functions_to_use_in_callback_validate_email = (
        functions_to_validate_email if functions_to_validate_email else DEFAULT_VALIDATORS_FUNCTION
    )
    EVEREST_LIST_INVALIDS_STATUS_RESPONSES = get_from_settings_or_raise_missing_config(
        "EVEREST_LIST_INVALIDS_STATUS_RESPONSES", "invalid"
    )
    invalid_status_list = EVEREST_LIST_INVALIDS_STATUS_RESPONSES.split(",")
    try:
        email_validation_response: Dict = get_email_validation(email)
        results: dict = email_validation_response.get("results") or {}
        status: Optional[str] = results.get("status")
        is_valid: bool = status not in invalid_status_list and status is not None
        return ValidationResult(is_valid=is_valid, external_api_validation=True)
    except Exception as err:
        logger.warning(f"Everest API Response Failed: {err}")
        if use_callback_function_validator:
            return ValidationResult(
                is_valid=perform_callback_function_validators(
                    functions_to_validate_email=functions_to_use_in_callback_validate_email, email=email
                ),
                external_api_validation=False,
            )
        raise EmailValidationError


def get_email_validation(email: str) -> Dict:
    """
    Returns the payload from Everest
    """
    EVEREST_API_KEY = get_from_settings_or_raise_missing_config("EVEREST_API_KEY", "api-key-everest")
    EVEREST_API_URL = get_from_settings_or_raise_missing_config(
        "EVEREST_API_HOST", "https://api.everest.validity.com/api/2.0/validation/addresses"
    )

    with request(status_forcelist=(429, 500, 502, 503, 504)) as r:
        url: str = f"{EVEREST_API_URL}/{email}"
        headers: Dict = {"X-API-KEY": EVEREST_API_KEY}
        response: Response = r.get(url, headers=headers)
        payload: Dict = get_response_body(response)
    logger.info(f"Everest API Response: {payload}")
    return payload



import logging
import os

from typing import Optional

from django.http import HttpRequest

from jsm_user_services.services.user import get_session_id_from_bearer_token
from jsm_user_services.support.auth_jwt import get_bearer_authorization_token
from jsm_user_services.support.http_utils import convert_header_to_meta_key
from jsm_user_services.support.local_threading_utils import add_to_local_threading
from jsm_user_services.support.local_threading_utils import remove_from_local_threading
from jsm_user_services.support.string_utils import get_first_value_from_comma_separated_string

logger = logging.getLogger(__name__)

header_with_client_ip_address = convert_header_to_meta_key(os.getenv("PRIMARY_IP_ADDRESS_HEADER", "true-client-ip"))
header_with_list_of_addresses = convert_header_to_meta_key(
    os.getenv("GUNICORN_IP_ADDRESS_HEADER", "x-original-forwarded-for")
)


class JsmJwtService:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        self.process_request(request)
        response = self.get_response(request)
        return self.process_response(request, response)

    def process_request(self, request):
        add_to_local_threading("authorization_token", self._get_jwt_token_from_request(request))
        add_to_local_threading("user_ip", self._get_user_ip_from_request(request))
        add_to_local_threading("user_session_id", self._get_user_session_id_from_request(request))

    def process_response(self, request, response):
        remove_from_local_threading("authorization_token")
        remove_from_local_threading("user_ip")
        return response

    @staticmethod
    def _get_jwt_token_from_request(request: HttpRequest) -> Optional[str]:
        """
        Extracts JWT token from a Django request object.
        """
        authorization_value = request.META.get("HTTP_AUTHORIZATION", "")
        return get_bearer_authorization_token(authorization_value)

    @staticmethod
    def _get_user_ip_from_request(request: HttpRequest) -> Optional[str]:
        """
        Retrieve the user ip that made this request from Django HttpRequest object

        When running a service behind Akamai or other CDN solutions, it is expected that this header might contain
        a string with multiple IPs (comma separated values).
        In this case, the user's public IP that originated the request is considered to be the first one of this list.
        For reference: https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/X-Forwarded-For
        """
        first_attempt_value = request.META.get(header_with_client_ip_address)
        if first_attempt_value:
            return first_attempt_value.strip()

        second_attempt_value = request.META.get(header_with_list_of_addresses)

        if not second_attempt_value:
            logger.warning("No user IP was detected!")
            return None

        return get_first_value_from_comma_separated_string(second_attempt_value)

    @staticmethod
    def _get_user_session_id_from_request(request: HttpRequest) -> Optional[str]:
        bearer_token = JsmJwtService._get_jwt_token_from_request(request)
        if not bearer_token:
            return None
        return get_session_id_from_bearer_token(bearer_token)



from rest_framework import status
from rest_framework.exceptions import APIException


class InvalidToken(APIException):
    status_code = status.HTTP_401_UNAUTHORIZED
    default_detail = "Token is invalid or expired"
    default_code = "TOKEN_NOT_VALID"


class NotAuthenticated(APIException):
    status_code = status.HTTP_401_UNAUTHORIZED
    default_detail = "Authentication credentials were not provided."
    default_code = "NOT_AUTHENTICATED"



from importlib import import_module
from typing import Callable

from jsm_user_services.exception import RequestIDModuleNotFound
from jsm_user_services.support.import_utils import import_module_otherwise_none


def parse_meta_header_to_request_header(value: str) -> str:
    return value.replace("HTTP_", "").replace("_", "-")


def get_request_id_header_name() -> str:
    """
    Helper function to get request_id header name to included in request to another application
    REQUEST_ID_CONFIG dict configuration exists in the projects with similarity structure below
    REQUEST_ID_CONFIG = {
        "REQUEST_ID_HEADER": "HTTP_X_REQUEST_ID",
        "GENERATE_REQUEST_ID_IF_NOT_FOUND": True,
        "RESPONSE_HEADER_REQUEST_ID": "HTTP_X_REQUEST_ID",
    }
    """
    settings = import_module("jsm_user_services.settings")
    default_request_id_header_name = getattr(settings, "REQUEST_ID_HEADER_NAME", "HTTP_X_REQUEST_ID")
    default_request_id_config = {
        "GENERATE_REQUEST_ID_IF_NOT_FOUND": True,
        "REQUEST_ID_HEADER": default_request_id_header_name,
        "RESPONSE_HEADER_REQUEST_ID": default_request_id_header_name,
    }
    request_id_config = getattr(settings, "REQUEST_ID_CONFIG", default_request_id_config)
    if request_id_config and isinstance(request_id_config, dict) and request_id_config.get("REQUEST_ID_HEADER"):
        return parse_meta_header_to_request_header(
            request_id_config.get("REQUEST_ID_HEADER", default_request_id_header_name)
        )
    return parse_meta_header_to_request_header(default_request_id_header_name)


def get_current_request_id_callable() -> Callable:
    django_request_id_module = import_module_otherwise_none("request_id_django_log.request_id")

    if django_request_id_module is not None:
        return django_request_id_module.current_request_id

    raise RequestIDModuleNotFound


current_request_id = get_current_request_id_callable()
request_id_header_name = get_request_id_header_name()



import time

import jwt

from django.contrib.auth.models import AnonymousUser
from rest_framework import authentication

from jsm_user_services.drf_tools.exceptions import InvalidToken as CustomInvalidToken
from jsm_user_services.drf_tools.exceptions import NotAuthenticated as CustomNotAuthenticated
from jsm_user_services.services.user import current_jwt_token
from jsm_user_services.support.auth_jwt import get_decoded_oauth_token


class OauthJWTAuthentication(authentication.BaseAuthentication):
    def authenticate(self, request):
        token = current_jwt_token()
        if token is None:
            raise CustomNotAuthenticated()

        try:
            payload = get_decoded_oauth_token(token)
            current_timestamp = int(time.time())
            is_token_expired = "exp" in payload and current_timestamp > payload["exp"]
            is_sub_claim_in_payload = "sub" in payload
            if not is_token_expired and is_sub_claim_in_payload:
                return (AnonymousUser(), token)
            else:
                raise CustomInvalidToken()
        except jwt.DecodeError:
            raise CustomInvalidToken()



from typing import Optional


def get_first_value_from_comma_separated_string(comma_separated_string: str) -> Optional[str]:
    if comma_separated_string:
        return comma_separated_string.split(sep=",")[0].strip()

    return None



"""
Module with useful user permissions classes meant to be used with Django Rest Framework.
"""

import logging

from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Type

import requests

from django.conf import settings
from rest_framework import permissions
from rest_framework.exceptions import APIException
from rest_framework.request import Request

from jsm_user_services.drf_tools.helpers import AllowClassBehaviorAsFunction
from jsm_user_services.drf_tools.helpers import is_exception_related_to_api_exception
from jsm_user_services.exception import IncorrectTypePermissionConfiguration
from jsm_user_services.services.google import perform_recaptcha_validation
from jsm_user_services.services.user import get_jwt_algorithm
from jsm_user_services.services.user import get_user_data_from_server
from jsm_user_services.services.user import get_user_id_auth
from jsm_user_services.services.user import get_user_id_auth_or_user_id_ref_in_jwt

logger = logging.getLogger(__name__)


class JSMUserBasePermission(permissions.BasePermission):
    """
    Base class for JSM user permissions. Implements methods for validating requests against
    the user micro service so that any type of permission can be carried out (role, status, etc).
    """

    APPEND_USER_DATA = getattr(settings, "JSM_USER_SERVICES_DRF_APPEND_USER_DATA", True)
    USER_DATA_ATTR_NAME = getattr(settings, "JSM_USER_SERVICES_DRF_REQUEST_USER_DATA_ATTR_NAME", "user_data")
    USER_SERVICE_NOT_AUTHORIZED_CODES = (401, 403, 404)

    @classmethod
    def _retrieve_user_data(cls, request: Request) -> Dict:
        try:
            user_data = getattr(request, cls.USER_DATA_ATTR_NAME)
        except AttributeError:
            user_data = get_user_data_from_server()

        return user_data

    @classmethod
    def _validate_request_against_user_service(cls, request: Request, append_user_data_to_request: bool = True) -> Dict:
        """
        Gets valid user_data from the User micro service.
        """
        try:
            user_data: Dict = cls._retrieve_user_data(request)

            if append_user_data_to_request:
                setattr(request, cls.USER_DATA_ATTR_NAME, user_data)

        except requests.exceptions.HTTPError as e:
            if e.response.status_code in cls.USER_SERVICE_NOT_AUTHORIZED_CODES:
                return {}

            raise

        return user_data

    def has_permission(self, request: Request, view: Any) -> bool:
        """
        Abstract method that must be implemented by children classes.
        """
        raise NotImplementedError("This method must be overridden!")


class StatusBasedPermission(JSMUserBasePermission):
    """
    Base class for the status-based permissions. Implements methods for validating requests against
    the user micro service and an utility method that asserts that the user has the appropriate status.
    """

    @classmethod
    def _validate_user_status(cls, request: Request, allowed_status: List[str]):
        """
        Validates if the user has the appropriate status against the allowed status.
        """
        append_user_data_to_request = cls.APPEND_USER_DATA
        user_data = cls._validate_request_against_user_service(
            request, append_user_data_to_request=append_user_data_to_request
        )

        if "status" not in user_data:
            return False

        return user_data["status"] in allowed_status

    @classmethod
    def _validate_user_blocked_reason(cls, request: Request, blocked_reasons: List[str]):
        """
        Validates if the user has the appropriate blocked reason against the allowed blocked reasons.
        """
        append_user_data_to_request = cls.APPEND_USER_DATA
        user_data = cls._validate_request_against_user_service(
            request, append_user_data_to_request=append_user_data_to_request
        )

        if not user_data.get("blocked_reason") or not user_data["blocked_reason"].get("reason"):
            return False

        return user_data["blocked_reason"]["reason"] in blocked_reasons


class ActiveUserPermission(StatusBasedPermission):
    """
    Permission that allows only users which the status is 'active'.
    """

    def has_permission(self, request: Request, view: Any) -> bool:
        """
        Asserts that the user status is 'active'.
        """
        return self._validate_user_status(request, allowed_status=["active"])


class PendingValidationUserPermission(StatusBasedPermission):
    """
    Permission that allows only users which the status is 'pending-validation'.
    """

    def has_permission(self, request: Request, view: Any) -> bool:
        """
        Asserts that the user status is 'pending-validation'.
        """
        return self._validate_user_status(request, allowed_status=["pending-validation"])


class UserBlockedByCpfAndBirthDateValidationFailedPermission(StatusBasedPermission):
    """
    Permission that allows only users which the status is 'blocked' and
    blocked reason is 'cpf-birth-date-validation-failed'.
    """

    def has_permission(self, request: Request, view: Any) -> bool:
        """
        Asserts that the user status is 'blocked' and blocked reason is 'cpf-birth-date-validation-failed'.
        """
        return self._validate_user_status(request, allowed_status=["blocked"]) and self._validate_user_blocked_reason(
            request, blocked_reasons=["cpf-birth-date-validation-failed"]
        )


class RoleBasedPermission(JSMUserBasePermission):
    """
    Base class for the role-based permissions. Implements methods for validating requests against
    the user micro service and an utility method that asserts that the user has the appropriate role.
    """

    @classmethod
    def _validate_user_role(cls, request: Request, allowed_roles: List[str]):
        """
        Validates if the user has the appropriate role against the allowed roles.
        """
        append_user_data_to_request = cls.APPEND_USER_DATA
        user_data = cls._validate_request_against_user_service(
            request, append_user_data_to_request=append_user_data_to_request
        )
        user_roles: List[str] = user_data.get("roles", [])
        if not user_roles:
            return False

        return any([user_role in allowed_roles for user_role in user_roles])


class RetailUserPermission(RoleBasedPermission, ActiveUserPermission):
    """
    Permission that allows only active users with the roles 'owner', 'manager' or 'employee'.
    """

    def has_permission(self, request: Request, view: Any) -> bool:
        """
        Asserts that the user is an active owner, manager or employee.
        """
        is_user_active = ActiveUserPermission.has_permission(self, request, view)
        has_retail_role = self._validate_user_role(request, allowed_roles=["owner", "employee", "manager"])
        return is_user_active and has_retail_role


class EmployeeOrManagerUserPermission(RoleBasedPermission, ActiveUserPermission):
    """
    Permission that allows only active users with the role 'employee' or 'manager'.
    """

    def has_permission(self, request: Request, view: Any) -> bool:
        """
        Asserts that the user is an active employee or manager.
        """
        is_user_active = ActiveUserPermission.has_permission(self, request, view)
        has_employee_or_manager_role = self._validate_user_role(request, allowed_roles=["employee", "manager"])
        return is_user_active and has_employee_or_manager_role


class OwnerOrManagerUserPermission(RoleBasedPermission, ActiveUserPermission):
    """
    Permission that allows only active users with the role 'owner' or 'manager'.
    """

    def has_permission(self, request: Request, view: Any) -> bool:
        """
        Asserts that the user is an active owner or manager.
        """
        is_user_active = ActiveUserPermission.has_permission(self, request, view)
        has_owner_or_manager_role = self._validate_user_role(request, allowed_roles=["manager", "owner"])
        return is_user_active and has_owner_or_manager_role


class ManagerUserPermission(RoleBasedPermission, ActiveUserPermission):
    """
    Permission that allows only active users with the role 'manager'.
    """

    def has_permission(self, request: Request, view: Any) -> bool:
        """
        Asserts that the user is an active manager.
        """
        is_user_active = ActiveUserPermission.has_permission(self, request, view)
        has_manager_role = self._validate_user_role(request, allowed_roles=["manager"])
        return is_user_active and has_manager_role


class EmployeeUserPermission(RoleBasedPermission, ActiveUserPermission):
    """
    Permission that allows only active users with the role 'employee'.
    """

    def has_permission(self, request: Request, view: Any) -> bool:
        """
        Asserts that the user is an active employee.
        """
        is_user_active = ActiveUserPermission.has_permission(self, request, view)
        has_employee_role = self._validate_user_role(request, allowed_roles=["employee"])
        return is_user_active and has_employee_role


class OwnerUserPermission(RoleBasedPermission, ActiveUserPermission):
    """
    Permission that allows only active users with the role 'owner'.
    """

    def has_permission(self, request: Request, view: Any) -> bool:
        """
        Asserts that the user is an active owner.
        """
        is_user_active = ActiveUserPermission.has_permission(self, request, view)
        has_owner_role = self._validate_user_role(request, allowed_roles=["owner"])
        return is_user_active and has_owner_role


class IndustrySellerUserPermission(RoleBasedPermission, ActiveUserPermission):
    """
    Permission that allows only active users with the role 'industry-seller'.
    """

    def has_permission(self, request: Request, view: Any) -> bool:
        """
        Asserts that the user is an active industry seller.
        """
        is_user_active = ActiveUserPermission.has_permission(self, request, view)
        has_industry_seller_role = self._validate_user_role(request, allowed_roles=["industry-seller"])
        return is_user_active and has_industry_seller_role


class AnyLoggedUserPermission(RoleBasedPermission, ActiveUserPermission):
    """
    Permission that allows active users to access the requested resource regardless
    of their roles as long as they have a VALID, NON EXPIRED JSM JWT.
    """

    def has_permission(self, request: Request, view: Any) -> bool:
        """
        Asserts that the active user is just logged in with a valid jwt (non expired).
        """
        is_user_active = ActiveUserPermission.has_permission(self, request, view)
        has_valid_role = self._validate_user_role(
            request, allowed_roles=["industry-seller", "employee", "manager", "owner"]
        )
        return is_user_active and has_valid_role


class GoogleRecaptchaPermission(permissions.BasePermission, AllowClassBehaviorAsFunction):
    """
    A permission class which checks if the request is authorized by Google Recaptcha V3.

    In order to use it, the request must have the key "g_recaptcha_response" or "g-recaptcha-response" on the request's
    body or header. Otherwise the request won't be authorized without even checking it on Google.

    Just be aware that a header with underscore is not allowed. So do not use the key "g_recaptcha_response"
    on the header.
    """

    def __init__(self, exception_in_case_of_failed: Optional[APIException | Type[APIException]] = None):
        """
        Sets the default value for "exception_in_case_of_failed_verification" property. It will be used to return
        a custom response, in case of this permission fails.
        This exception must inherit APIException, from rest_framework.exceptions, otherwise it will be ignored.
        """

        self.exception_in_case_of_failed_verification = exception_in_case_of_failed

    def has_permission(self, request: Request, view: Any) -> bool:
        """
        Retrieves the Google data on the request and then performs a request to Google, in order to check if the
        received data is valid or not.
        """

        g_recaptcha_response = self._retrieve_g_recaptcha_response(request)
        use_received_exception = is_exception_related_to_api_exception(self.exception_in_case_of_failed_verification)

        # preventing an unnecessary request to Google, since key is not defined
        if not g_recaptcha_response:
            if use_received_exception:
                raise self.exception_in_case_of_failed_verification  # type: ignore
            return False

        google_response: bool = perform_recaptcha_validation(g_recaptcha_response)

        if not google_response and use_received_exception:
            # raising an APIException that will return the customized response
            raise self.exception_in_case_of_failed_verification  # type: ignore

        return google_response

    @classmethod
    def _retrieve_g_recaptcha_response(cls, request: Request) -> Optional[str]:
        """
        Lookup on the request's body and header in order to retrieve the recaptcha key.

         If the same key is present on header and body, the header data will be overridden by the body value.
         Also, if both keys are present, the priority is "g_recaptcha_response" over "g-recaptcha-response".
        """

        expected_keys = ["g_recaptcha_response", "g-recaptcha-response"]

        for key in expected_keys:
            key_value = request.data.get(key) or request.headers.get(key)
            if key_value:
                return key_value
        return None


class IsUserAllowedToPermission(JSMUserBasePermission, AllowClassBehaviorAsFunction):
    """
    A permission class which checks if user is authorized to execute an action based on 'not_allowed_to' parameter.
    Implements methods for validating requests against the user micro service and an utility method that asserts
    the user has the appropriate permission for to do, or not to do, a determined action.

    Example:
        class SomeAwesomeView():
            permission_classes = [
                IsUserAllowedToPermission(
                    ["redeem", "score"], full_match=False, exception_in_case_of_failed=CustomExceptionNotAllowedTo
                )
            ]

        In this example, the custom exception should be implemented into your own project.
    """

    def __init__(
        self,
        not_allowed_to: List[str],
        full_match: bool = False,
        exception_in_case_of_failed: Optional[APIException | Type[APIException]] = None,
    ):
        """
        Sets the default value for "exception_in_case_of_failed_verification" property. It will be used to return
        a custom response, in case of this permission fails. This exception must inherit APIException,
        from rest_framework.exceptions, otherwise it will be ignored.

        Full match property indicates that user data must contains all the 'not_allowed_to' defined parameters.
        On the other hand, if full match is equal to False, the permission class checks
        if user data contains any of the 'not_allowed_to' parameters.
        """

        self.exception_in_case_of_failed_verification = exception_in_case_of_failed
        self.full_match = full_match

        if not isinstance(not_allowed_to, list):
            logger.error(
                f"Permission class 'IsUserAllowedToPermission' expected 'List[str]', got  {type(not_allowed_to)}"
            )
            raise IncorrectTypePermissionConfiguration(f"Expected type 'List[str]', got {type(not_allowed_to)} instead")

        self.not_allowed_to = not_allowed_to

    @classmethod
    def _verify_if_user_should_be_allowed(
        cls, request: Request, not_allowed_actions: List[str], full_match: bool
    ) -> bool:
        """
        Validates 'not_allowed_to' user parameter against the 'not_allowed_actions' permission class parameter.
        """

        append_user_data_to_request = cls.APPEND_USER_DATA
        user_data = cls._validate_request_against_user_service(
            request, append_user_data_to_request=append_user_data_to_request
        )
        user_not_allowed_to: List[str] = user_data.get("not_allowed_to", [])

        set_from_not_allowed_actions = set(not_allowed_actions)
        set_from_user_not_allowed_to = set(user_not_allowed_to)
        intersected_actions = set_from_not_allowed_actions.intersection(set_from_user_not_allowed_to)

        if not (user_not_allowed_to or intersected_actions):
            return True

        if full_match:
            return len(intersected_actions) != len(set_from_not_allowed_actions)
        return not bool(intersected_actions)

    def has_permission(self, request: Request, view: Any) -> bool:
        """
        Asserts that the user hasn't an action restriction against the 'not_allowed_to' permission class parameter.
        """

        use_received_exception: bool = is_exception_related_to_api_exception(
            self.exception_in_case_of_failed_verification
        )

        is_allowed: bool = self._verify_if_user_should_be_allowed(
            request,
            not_allowed_actions=self.not_allowed_to,
            full_match=self.full_match,
        )

        if use_received_exception and not is_allowed:
            raise self.exception_in_case_of_failed_verification  # type: ignore

        return is_allowed


class LoggedUserOauthAndLegacyPermission(JSMUserBasePermission):
    """
    Permission to check if the user is authenticated with an Oauth or Loyalty JWT.
    """

    def has_permission(self, request, *args, **kwargs):
        alg = get_jwt_algorithm()
        if alg == "RS256":
            user = get_user_id_auth_or_user_id_ref_in_jwt()
            return bool(user)
        else:
            return bool(request.user and request.user.is_authenticated)


class OauthUserPermission(JSMUserBasePermission):
    """
    Permission to check if the user is authenticated with an Oauth JWT.
    """

    def has_permission(self, request, *args, **kwargs):
        user = get_user_id_auth()
        return bool(user)



from typing import Callable
from typing import Dict
from typing import Optional

from jsm_user_services.decorators.lgpd_utils import AnonymizeLogic
from jsm_user_services.decorators.lgpd_utils import traverse_within_dict_overriding_data
from jsm_user_services.exception import DecoratorWrongUsage


def anonymize_sensible_data(
    to_be_anonymized: Dict, raise_if_key_not_found: Optional[bool] = True, traverse_dict_pattern: Optional[str] = "__"
):
    """
    Method to be used as decorator that will anonymize sensible data, according to LGPD.

    In order to use it, the "original_method" MUST return a dict. Example:
    >>> @anonymize_sensible_data({"name": AnonymizeLogic.NAME})
    ... def honest_test():
    ...    return {"name": "Lucas Gigek Carvalho"}
    ...
    >>> honest_test()
        {"name": "Lucas C"}

    It also should be called with a dict that follows this pattern:
    {"field_that_must_be_anonymized": "anonimyze_logic_for_specific_field"}

    If the dict is more complex than the example above, use "__" to traverse within dicts/lists. Example:
    - For {"userdata": {"name": "xpto"}}, use "userdata__name";
    - For {"result": [{"userdata": {"name": "xpto"}, {"userdata": {"name": "xpto2"}]}, use "result__userdata__name";
    Example:
    >>> @anonymize_sensible_data({"userdata__name": AnonymizeLogic.NAME})
    ... def honest_test():
    ...    return {"userdata": {"name": "Lucas Gigek Carvalho"}}
    ...
    >>> honest_test()
        {"userdata": {"name": "Lucas C"}}

    It's possible to customize if it should raise or not if some key wasn't found via param "raise_if_key_not_found".
    It's possible to customize the pattern to traverse the dict as well. The default value is "__".
    """

    def wrapper(method_that_uses_the_decorator: Callable):
        """
        Since the decorator must be called with a parameter, this method is executed when the module is imported. It
        checks if "to_be_anonymized" is valid. A valid param must:
        - be a dict;
        - have keys that are a string;
        - have values that are AnonymizeLogic;
        If the criteria above is violated, raises DecoratorWrongUsage

        After that, simply return the main logic, which will be executed when the method is invoked.
        """

        if not isinstance(to_be_anonymized, dict):
            raise DecoratorWrongUsage("Please, pass a dict as param.")

        for key, value in to_be_anonymized.items():
            if not isinstance(key, str):
                raise DecoratorWrongUsage("The keys must be a str, got key %s of type %s", key, type(key))
            if not isinstance(value, Callable) or not hasattr(AnonymizeLogic, value.__name__):  # type: ignore
                raise DecoratorWrongUsage("The values must be a AnonymizeLogic, got value %s", value)

        def main_logic(*args, **kwargs):
            """
            Method that will be executed only when the method with decorator is called. Given a field to be
            anonymized (key) and the logic that it should use (value), it anonymizes the dict to be in agreement with
            LGPD.
            """

            original_result = method_that_uses_the_decorator(*args, **kwargs)

            if not isinstance(original_result, dict) and not isinstance(original_result, list):
                raise DecoratorWrongUsage(
                    "The method must return a dict or a list, but returned type %s", type(original_result)
                )

            for path, strategy in to_be_anonymized.items():
                traverse_within_dict_overriding_data(
                    path, original_result, strategy, raise_if_key_not_found, traverse_dict_pattern
                )

            return original_result

        return main_logic

    return wrapper



import hashlib

from importlib import import_module
from typing import Any
from typing import List
from typing import Optional

import jwt

from jsm_user_services.support.auth_jwt import get_decoded_jwt_token
from jsm_user_services.support.auth_jwt import get_decoded_oauth_token
from jsm_user_services.support.http_utils import get_response_body
from jsm_user_services.support.http_utils import request
from jsm_user_services.support.local_threading_utils import get_from_local_threading
from jsm_user_services.support.request_id import current_request_id
from jsm_user_services.support.request_id import request_id_header_name


def jwt_has_required_roles(jwt_required_roles: List[str], assert_all=True) -> bool:
    """
    Asserts that the jwt token has all the required roles.
    """
    jsm_token_data = get_jsm_user_data_from_jwt()

    if not jsm_token_data:
        return False

    try:
        jwt_token_roles = jsm_token_data["roles"]
    except KeyError:
        return False

    if not jwt_token_roles:
        return False

    if assert_all:
        return all([(jwt_token_role in jwt_required_roles) for jwt_token_role in jwt_token_roles])

    return any([(jwt_token_role in jwt_required_roles) for jwt_token_role in jwt_token_roles])


def current_jwt_token() -> Optional[str]:
    return get_from_local_threading("authorization_token")


def get_jwt_algorithm() -> str:
    """
    Retrieves the algorithm used in the JWT token.
    """
    token = current_jwt_token()
    header = jwt.get_unverified_header(token)
    return header.get("alg")


def get_jsm_token() -> Optional[str]:
    token = current_jwt_token()
    if token:
        return get_decoded_jwt_token(token)["jsm_identity"]

    return None


def get_oauth_token() -> Optional[dict[Any, Any]]:
    """
    This function retrieves the OAuth token from the local threading.
    If the token is not present, it returns None.
    """
    token = current_jwt_token()
    if token:
        return get_decoded_oauth_token(token)
    return None


def get_ltm_token() -> Optional[str]:
    token = current_jwt_token()
    if token:
        return get_decoded_jwt_token(token)["yuntiandu"]

    return None


def get_jsm_user_data_from_jwt() -> Optional[dict]:
    token = get_jsm_token()

    if token:
        return get_decoded_jwt_token(token)

    return None


def get_user_email_from_jwt() -> Optional[str]:
    user_data = get_jsm_user_data_from_jwt()
    if user_data:
        return user_data.get("email")

    return None


def get_user_id_from_jwt() -> Optional[str]:
    user_data = get_jsm_user_data_from_jwt()
    if user_data:
        return user_data.get("uid")

    return None


def get_user_access_as_id_from_jwt() -> Optional[str]:
    user_data = get_jsm_user_data_from_jwt()
    if user_data:
        return user_data.get("fid", None)

    return None


def get_user_data_from_server() -> dict:
    current_token = current_jwt_token()
    headers = {request_id_header_name: current_request_id()}
    settings = import_module("jsm_user_services.settings")
    user_url = settings.USER_API_HOST

    if current_token:
        headers["Authorization"] = f"Bearer {current_jwt_token()}"

    with request() as r:
        response = r.get(f"{user_url}/users/me/", headers=headers)
        return get_response_body(response)


def get_user_data_from_cpf(cpf: str) -> dict:
    current_token = current_jwt_token()
    headers = {request_id_header_name: current_request_id()}
    settings = import_module("jsm_user_services.settings")
    user_url = settings.USER_API_HOST

    if current_token:
        headers["Authorization"] = f"Bearer {current_jwt_token()}"

    with request() as r:
        response = r.get(f"{user_url}/users/search?cpf={cpf}", headers=headers)
        return get_response_body(response)


def get_user_data_from_id(user_id: str) -> dict:
    current_token = current_jwt_token()
    headers = {request_id_header_name: current_request_id()}
    settings = import_module("jsm_user_services.settings")
    user_url = settings.USER_API_HOST

    if current_token:
        headers["Authorization"] = f"Bearer {current_jwt_token()}"

    with request() as r:
        response = r.get(f"{user_url}/users/{user_id}/", headers=headers)

        return get_response_body(response)


def get_cpf_from_jwt() -> Optional[str]:
    email = get_user_email_from_jwt()

    return email.split("@")[0] if email else None


def is_retail_user(user_id: str) -> bool:
    settings = import_module("jsm_user_services.settings")
    user_api_token = settings.USER_API_TOKEN
    user_url = settings.USER_API_HOST

    headers = {request_id_header_name: current_request_id(), "Authorization": f"Token {user_api_token}"}

    with request() as r:
        response = r.get(f"{user_url}/users/search/?user_id_ref={user_id}&is_retail_user=True", headers=headers)
        response_content = get_response_body(response)
    return response_content["count"] == 1


def get_user_ip() -> Optional[str]:
    return get_from_local_threading("user_ip")


def get_session_id_from_bearer_token(bearer_token: str) -> str:
    return hashlib.sha1(bearer_token.encode("utf-8")).hexdigest()


def get_user_session_id() -> Optional[str]:
    return get_from_local_threading("user_session_id")


def get_user_id_auth() -> Optional[str]:
    """
    This function retrieves the user's ID from an OAuth JWT token.
    If the "sub" field is not present, it returns None.
     Note:
        If the claim "https://login.juntossomosmais.com.br/user_id_auth" is present in the JWT from Auth0, it indicates
        that the main information in the token belongs to the Central user (not the actual user).
        Therefore, we must return the user_id_auth, as this is the ID that represents the real user.
    """
    user_data = get_oauth_token()

    if not user_data:
        return None

    return user_data.get("https://login.juntossomosmais.com.br/user_id_auth") or user_data.get("sub")


def get_user_email_from_oauth_jwt() -> Optional[str]:
    """
    This function retrieves the user's email from an OAuth JWT token.
    If the "email" field is not present, it returns None.

    Note:
        If the claim "https://login.juntossomosmais.com.br/user_id_auth" is present in the JWT from Auth0, it indicates
        that the main information in the token belongs to the Central user (not the actual user).
        Therefore, we must return None in this case, as the email does not represent the real user.

    """
    user_data = get_oauth_token()

    if not user_data or user_data.get("https://login.juntossomosmais.com.br/user_id_auth"):
        return None

    return user_data.get("email")


def get_user_id_auth_or_user_id_ref_in_jwt() -> Optional[str]:
    """
    Retrieves the user_id_auth or user_id_ref from the JWT token according to the algorithm used.
    """
    alg = get_jwt_algorithm()

    alg_to_user_id_map = {"RS256": get_user_id_auth, "HS256": get_user_id_from_jwt}

    return alg_to_user_id_map.get(alg, lambda: None)()



from typing import Optional

from jsm_user_services.exception import ApplicationSettingsNotFound
from jsm_user_services.support.import_utils import import_module_otherwise_none


def get_django_settings():
    django_conf = import_module_otherwise_none("django.conf")
    if django_conf and django_conf.settings.configured:
        return django_conf.settings
    return None


def get_setting_from_app(setting_name: str, default_value: Optional[str] = None):
    django_settings = get_django_settings()
    if django_settings is not None:
        return getattr(django_settings, setting_name, default_value)
    raise ApplicationSettingsNotFound


USER_API_HOST = get_setting_from_app("USER_API_HOST")
USER_API_TOKEN = get_setting_from_app("USER_API_TOKEN")
JSM_USER_SERVICE_HTTP_TIMEOUT = get_setting_from_app("JSM_USER_SERVICE_HTTP_TIMEOUT", "30")
GOOGLE_RECAPTCHA_URL = get_setting_from_app("GOOGLE_RECAPTCHA_URL")
GOOGLE_RECAPTCHA_SECRET_KEY = get_setting_from_app("GOOGLE_RECAPTCHA_SECRET_KEY")
GOOGLE_RECAPTCHA_SCORE_THRESHOLD = get_setting_from_app("GOOGLE_RECAPTCHA_SCORE_THRESHOLD")
GOOGLE_RECAPTCHA_BYPASS_IP_HEADER_NAME = get_setting_from_app("GOOGLE_RECAPTCHA_BYPASS_IP_HEADER_NAME")
EVEREST_API_KEY = get_setting_from_app("EVEREST_API_KEY")
EVEREST_API_HOST = get_setting_from_app("EVEREST_API_HOST")
EVEREST_LIST_INVALIDS_STATUS_RESPONSES = get_setting_from_app("EVEREST_LIST_INVALIDS_STATUS_RESPONSES")
REQUEST_ID_CONFIG = get_setting_from_app("REQUEST_ID_CONFIG")



import logging

from jsm_user_services.services.user import get_user_session_id


class UserSessionIDFilter(logging.Filter):
    EMPTY_FIELD = "-"

    def filter(self, record):
        record.session_id = get_user_session_id() or self.EMPTY_FIELD
        return True



from contextlib import suppress
from importlib import import_module
from types import ModuleType
from typing import Optional


def import_module_otherwise_none(module_to_import: str) -> Optional[ModuleType]:
    with suppress(ImportError):
        return import_module(module_to_import)

    return None
